import 'package:node_interop/buffer.dart';
import 'package:node_interop/js.dart';
import 'package:node_interop/stream.dart';

import '../utils/js.dart';
import '../utils/utils.dart';
import 'interop/storage_interop.dart' as interop;

export 'interop/storage_interop.dart'
    show
        Acl,
        AddLifecycleRuleOptions,
        StorageBasicOptions,
        BucketExistsOptions,
        CombineOptions,
        CopyOptions,
        ContentLenghtRange,
        PolicyDocument,
        CreateReadStreamOptions,
        CreateChannelConfig,
        DeleteFilesOptions,
        FileOptions,
        GetBucketOptions,
        GetBucketsRequest,
        GetConfig,
        GetFilesOptions,
        Iam,
        LifeCycleRuleAction,
        MakeFilePrivateOptions,
        MakeAllFilesPublicPrivateOptions,
        MakeBucketPrivateOptions,
        MakeBucketPublicOptions,
        UrlSigner,
        Versioning;

Map<String, dynamic> _dartify(List list) => dartify(list.first);

Future<StorageResponse<T>> _makeResponse<T>(
  Future<List> future,
  T Function(dynamic o) mapper,
) async {
  final result = await future;
  final data = mapper(result[0]);
  final map = dartify(result[1]);
  return StorageResponse(data, map);
}

typedef GetFilesCallback = void Function([
  JsError error,
  List<File> files,
  Map<String, dynamic> nextQuery,
  Map<String, dynamic> apiResponse,
]);

class Bucket extends JsObjectWrapper<interop.BucketJsImpl> implements _Bucket {
  static final _expando = Expando<Bucket>();

  factory Bucket.getInstance(interop.BucketJsImpl jsObject) {
    if (jsObject == null) return null;
    return _expando[jsObject] ??= Bucket._fromJsObject(jsObject);
  }

  Bucket._fromJsObject(interop.BucketJsImpl jsObject)
      : super.fromJsObject(jsObject);

  @override
  Future<Map<String, dynamic>> addLifecycleRule(LifeCycleRule rule,
      [interop.AddLifecycleRuleOptions options]) {
    // TODO: implement addLifecycleRule
    throw UnimplementedError();
  }

  @override
  Future<StorageResponse<File>> combine(
    List /*String|File*/ sources,
    dynamic /*String,File*/ destination, [
    interop.CombineOptions options,
  ]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(
        jsObject.combine(
          jsify(sources),
          jsify(destination),
          options,
        ),
      );
    } else {
      future = handleThenable(
        jsObject.combine(
          jsify(sources),
          jsify(destination),
        ),
      );
    }
    return _makeResponse(future, (o) => File.getInstance(o));
  }

  @override
  Future<StorageResponse<Bucket>> create([CreateBucketRequest metadata]) {
    Future<List> future;
    if (metadata != null) {
      future = handleThenable(jsObject.create(metadata.jsObject));
    } else {
      future = handleThenable(jsObject.create());
    }
    return _makeResponse(future, (o) => Bucket.getInstance(o));
  }

  @override
  Future<StorageResponse<Channel>> createChannel(
    String id,
    interop.CreateChannelConfig config, [
    interop.StorageBasicOptions options,
  ]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.createChannel(id, config, options));
    } else {
      future = handleThenable(jsObject.createChannel(id, config));
    }
    return _makeResponse(future, (o) => Channel.getInstance(o));
  }

  @override
  Future<StorageResponse<StorageNotification>> createNotification(
    String topic, [
    CreateNotificationOptions options,
  ]) {
    Future<List> future;

    if (options != null) {
      future =
          handleThenable(jsObject.createNotification(topic, options.jsObject));
    } else {
      future = handleThenable(jsObject.createNotification(topic));
    }

    return _makeResponse(future, (o) => StorageNotification.getInstance(o));
  }

  @override
  Future<Map<String, dynamic>> delete(interop.StorageBasicOptions options) {
    // TODO: implement delete
    throw UnimplementedError();
  }

  @override
  Future<void> deleteFiles([interop.DeleteFilesOptions options]) {
    if (options != null) {
      return handleThenable(jsObject.deleteFiles(options));
    } else {
      return handleThenable(jsObject.deleteFiles());
    }
  }

  @override
  Future<Map<String, dynamic>> deleteLabels(List<String> labels) =>
      handleThenable(jsObject.deleteLabels(labels)).then(_dartify);

  @override
  Future<Map<String, dynamic>> disableRequesterPays() {
    return handleThenable(jsObject.disableRequesterPays())
        .then((value) => dartify(value?.first));
  }

  @override
  Future<Map<String, dynamic>> enableLogging(EnableLoggingOptions options) =>
      handleThenable(jsObject.enableLogging(options.jsObject))
          .then((value) => dartify(value?.first));

  @override
  Future<Map<String, dynamic>> enableRequesterPays() {
    return handleThenable(jsObject.enableRequesterPays())
        .then((value) => dartify(value?.first));
  }

  @override
  Future<bool> exists([interop.BucketExistsOptions options]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.exists(options));
    } else {
      future = handleThenable(jsObject.exists());
    }
    return future.then((value) => dartify(value.first));
  }

  @override
  File file(String name, [interop.FileOptions options]) {
    if (options != null) {
      return File.getInstance(jsObject.file(name, options));
    } else {
      return File.getInstance(jsObject.file(name));
    }
  }

  @override
  Future<StorageResponse<Bucket>> get([interop.GetBucketOptions options]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.get(options));
    } else {
      future = handleThenable(jsObject.get());
    }
    return _makeResponse(future, (o) => Bucket.getInstance(o));
  }

  @override
  Future<List<File>> getFiles([interop.GetFilesOptions options]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.getFiles(options));
    } else {
      future = handleThenable(jsObject.getFiles());
    }
    return future.then((value) =>
        (value.first as List).map((e) => File.getInstance(e)).toList());
  }

  @override
  Future<Map<String, dynamic>> getLabels(
      [interop.StorageBasicOptions options]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.getLabels(options));
    } else {
      future = handleThenable(jsObject.getLabels());
    }
    return future.then(_dartify);
  }

  @override
  Future<StorageResponse<Map<String, dynamic>>> getMetadata(
      [interop.StorageBasicOptions options]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.getMetadata(options));
    } else {
      future = handleThenable(jsObject.getMetadata());
    }
    return _makeResponse(future, (o) => dartify(o) as Map<String, dynamic>);
  }

  @override
  Future<StorageResponse<List<StorageNotification>>> getNotifications(
      [interop.StorageBasicOptions options]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.getNotifications(options));
    } else {
      future = handleThenable(jsObject.getNotifications());
    }

    return _makeResponse(future,
        (o) => (o as List).map((e) => StorageNotification.getInstance(e)));
  }

  @override
  Future<String> getSignedUrl(GetBucketSignedUrlConfig config) =>
      handleThenable(jsObject.getSignedUrl(config.jsObject))
          .then((value) => value.first as String);

  @override
  Future<Map<String, dynamic>> lock(dynamic /*num | String*/ metageneration) {
    assert(metageneration is num || metageneration is String);
    return handleThenable(jsObject.lock(metageneration)).then(_dartify);
  }

  @override
  Future<List<File>> makePrivate([interop.MakeBucketPrivateOptions options]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.makePrivate(options));
    } else {
      future = handleThenable(jsObject.makePrivate());
    }
    return future.then((value) =>
        (value.first as List).map((e) => File.getInstance(e)).toList());
  }

  @override
  Future<List<File>> makePublic([interop.MakeBucketPublicOptions options]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.makePublic(options));
    } else {
      future = handleThenable(jsObject.makePublic());
    }
    return future.then((value) =>
        (value.first as List).map((e) => File.getInstance(e)).toList());
  }

  @override
  StorageNotification notification(String id) =>
      StorageNotification.getInstance(jsObject.notification(id));

  @override
  Future<Map<String, dynamic>> removeRetentionPeriod() =>
      handleThenable(jsObject.removeRetentionPeriod()).then(_dartify);

  @override
  Future<Map<String, dynamic>> setLabels(Map<String, String> labels,
      [interop.StorageBasicOptions options]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.setLabels(jsify(labels), options));
    } else {
      future = handleThenable(jsObject.setLabels(jsify(labels)));
    }
    return future.then(_dartify);
  }

  @override
  Future<Map<String, dynamic>> setMetadata(Map<String, String> metadata,
      [interop.StorageBasicOptions options]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.setMetadata(jsify(metadata), options));
    } else {
      future = handleThenable(jsObject.setMetadata(jsify(metadata)));
    }
    return future.then(_dartify);
  }

  @override
  Future<Map<String, dynamic>> setRetentionPeriod(int duration) =>
      handleThenable(jsObject.setRetentionPeriod(duration)).then(_dartify);

  @override
  Future<Map<String, dynamic>> setStorageClass(String storageClass,
      [interop.StorageBasicOptions options]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.setStorageClass(storageClass, options));
    } else {
      future = handleThenable(jsObject.setStorageClass(storageClass));
    }
    return future.then(_dartify);
  }

  @override
  void setUserProject(String userProject) =>
      jsObject.setUserProject(userProject);

  @override
  Future<StorageResponse<File>> upload(String path, [UploadOptions options]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.upload(path, options.jsObject));
    } else {
      future = handleThenable(jsObject.upload(path));
    }
    return _makeResponse(future, (o) => File.getInstance(o));
  }
}

class Channel extends JsObjectWrapper<interop.ChannelJsImpl> {
  static final _expando = Expando<Channel>();

  factory Channel.getInstance(interop.ChannelJsImpl jsObject) {
    if (jsObject == null) return null;
    return _expando[jsObject] ??= Channel._fromJsObject(jsObject);
  }

  Channel._fromJsObject(interop.ChannelJsImpl jsObject)
      : super.fromJsObject(jsObject);

  Future<void> stop() => handleThenable(jsObject.stop());
}

class Cors extends JsObjectWrapper<interop.CorsJsImpl> {
  static final _expando = Expando<Cors>();

  factory Cors({
    num maxAgeSeconds,
    List<String> method,
    List<String> origin,
    List<String> responseHeader,
  }) =>
      Cors.getInstance(interop.CorsJsImpl(
        maxAgeSeconds: maxAgeSeconds,
        method: method,
        origin: origin,
        responseHeader: responseHeader,
      ));

  factory Cors.getInstance(interop.CorsJsImpl jsObject) {
    if (jsObject == null) return null;
    return _expando[jsObject] ??= Cors._fromJsObject(jsObject);
  }

  Cors._fromJsObject(interop.CorsJsImpl jsObject)
      : super.fromJsObject(jsObject);

  num get maxAgeSeconds => jsObject.maxAgeSeconds;
  set maxAgeSeconds(num v) => jsObject.maxAgeSeconds = v;

  List<String> get method => List.from(jsObject.method);
  set method(List<String> v) => jsObject.method = v;

  List<String> get origin => List.from(jsObject.origin);
  set origin(List<String> v) => jsObject.origin = v;

  List<String> get responseHeader => List.from(jsObject.responseHeader);
  set responseHeader(List<String> v) => jsObject.responseHeader = v;
}

class CreateBucketRequest
    extends JsObjectWrapper<interop.CreateBucketRequestJsImpl> {
  static final _expando = Expando<CreateBucketRequest>();

  factory CreateBucketRequest({
    bool archive,
    bool coldline,
    List<Cors> cors,
    bool dra,
    bool multiRegional,
    bool nearline,
    bool regional,
    bool requesterPays,
    bool standard,
    bool userProject,
    interop.Versioning versioning,
  }) =>
      CreateBucketRequest.getInstance(interop.CreateBucketRequestJsImpl(
        archive: archive,
        coldline: coldline,
        cors: jsify(cors),
        dra: dra,
        multiRegional: multiRegional,
        nearline: nearline,
        requesterPays: requesterPays,
        standard: standard,
        userProject: userProject,
        versioning: versioning,
      ));

  factory CreateBucketRequest.getInstance(
      interop.CreateBucketRequestJsImpl jsObject) {
    if (jsObject == null) return null;
    return _expando[jsObject] ??= CreateBucketRequest._fromJsObject(jsObject);
  }

  CreateBucketRequest._fromJsObject(interop.CreateBucketRequestJsImpl jsObject)
      : super.fromJsObject(jsObject);

  bool get archive => jsObject.archive;
  set archive(bool v) => jsObject.archive = v;

  bool get coldline => jsObject.coldline;
  set coldline(bool v) => jsObject.coldline = v;

  List<Cors> get cors => jsObject.cors.map((e) => Cors.getInstance(e)).toList();
  set cors(List<Cors> v) => jsObject.cors = v;

  bool get dra => jsObject.dra;
  set dra(bool v) => jsObject.dra = v;

  bool get multiRegional => jsObject.multiRegional;
  set multiRegional(bool v) => jsObject.multiRegional = v;

  bool get nearline => jsObject.nearline;
  set nearline(bool v) => jsObject.nearline = v;

  bool get regional => jsObject.regional;
  set regional(bool v) => jsObject.regional = v;

  bool get requesterPays => jsObject.requesterPays;
  set requesterPays(bool v) => jsObject.requesterPays = v;

  bool get standard => jsObject.standard;
  set standard(bool v) => jsObject.standard = v;

  bool get userProject => jsObject.userProject;
  set userProject(bool v) => jsObject.userProject = v;

  interop.Versioning get versioning => jsObject.versioning;
  set versioning(interop.Versioning v) => jsObject.versioning = v;
}

class CreateNotificationOptions
    extends JsObjectWrapper<interop.CreateNotificationOptionsJsImpl> {
  static final _expando = Expando<CreateNotificationOptions>();

  factory CreateNotificationOptions({
    Map<String, String> customAttributes,
    String eventTypes,
    String objectNamePrefix,
    String payloadFormat,
    String userProject,
  }) =>
      CreateNotificationOptions.getInstance(
          interop.CreateNotificationOptionsJsImpl(
        customAttributes: jsify(customAttributes),
        eventTypes: eventTypes,
        objectNamePrefix: objectNamePrefix,
        payloadFormat: payloadFormat,
        userProject: userProject,
      ));

  factory CreateNotificationOptions.getInstance(
      interop.CreateNotificationOptionsJsImpl jsObject) {
    if (jsObject == null) return null;
    return _expando[jsObject] ??=
        CreateNotificationOptions._fromJsObject(jsObject);
  }

  CreateNotificationOptions._fromJsObject(
      interop.CreateNotificationOptionsJsImpl jsObject)
      : super.fromJsObject(jsObject);

  Map<String, dynamic> get customAttributes {
    final map = dartify(jsObject.customAttributes);
    if (map == null) return null;
    return (map as Map<String, dynamic>).cast<String, String>();
  }

  set customAttributes(Map<String, dynamic> v) =>
      jsObject.customAttributes = jsify(v);

  String get eventTypes => jsObject.eventTypes;
  set eventTypes(String v) => jsObject.eventTypes = v;

  String get objectNamePrefix => jsObject.objectNamePrefix;
  set objectNamePrefix(String v) => jsObject.objectNamePrefix = v;

  String get payloadFormat => jsObject.payloadFormat;
  set payloadFormat(String v) => jsObject.payloadFormat = v;

  String get userProject => jsObject.userProject;
  set userProject(String v) => jsObject.userProject = v;
}

class CreateNotificationResponse {
  final StorageNotification notification;
  final Map<String, dynamic> object;
  CreateNotificationResponse(this.notification, this.object);
}

class CreateResumableUploadOptions
    extends JsObjectWrapper<interop.CreateResumableUploadOptionsJsImpl> {
  static final _expando = Expando<CreateResumableUploadOptions>();

  factory CreateResumableUploadOptions.getInstance(
      interop.CreateResumableUploadOptionsJsImpl jsObject) {
    if (jsObject == null) return null;
    return _expando[jsObject] ??=
        CreateResumableUploadOptions._fromJsObject(jsObject);
  }

  CreateResumableUploadOptions._fromJsObject(
      interop.CreateResumableUploadOptionsJsImpl jsObject)
      : super.fromJsObject(jsObject);
}

class CreateWriteStreamOptions
    extends JsObjectWrapper<interop.CreateWriteStreamOptionsJsImpl>
    implements _CreateWriteStreamOptions {
  static final _expando = Expando<CreateWriteStreamOptions>();

  factory CreateWriteStreamOptions.getInstance(
      interop.CreateWriteStreamOptionsJsImpl jsObject) {
    if (jsObject == null) return null;
    return _expando[jsObject] ??=
        CreateWriteStreamOptions._fromJsObject(jsObject);
  }

  CreateWriteStreamOptions._fromJsObject(
      interop.CreateWriteStreamOptionsJsImpl jsObject)
      : super.fromJsObject(jsObject);
}

class EnableLoggingOptions
    extends JsObjectWrapper<interop.EnableLoggingOptionsJsImpl> {
  static final _expando = Expando<EnableLoggingOptions>();

  factory EnableLoggingOptions({
    Map<String, Bucket> bucket,
    String prefix,
  }) =>
      EnableLoggingOptions.getInstance(interop.EnableLoggingOptionsJsImpl(
        bucket: jsify(bucket),
        prefix: prefix,
      ));

  factory EnableLoggingOptions.getInstance(
      interop.EnableLoggingOptionsJsImpl jsObject) {
    if (jsObject == null) return null;
    return _expando[jsObject] ??= EnableLoggingOptions._fromJsObject(jsObject);
  }

  EnableLoggingOptions._fromJsObject(
      interop.EnableLoggingOptionsJsImpl jsObject)
      : super.fromJsObject(jsObject);

  Map<String, Bucket> get bucket {
    final map = dartify(jsObject.bucket);
    if (map == null) return null;
    return map.map((k, v) => MapEntry(k as String, Bucket.getInstance(v)));
  }

  set Bucket(Map<String, Bucket> v) => jsObject.bucket = jsify(v);

  String get prefix => jsObject.prefix;
  set prefix(String v) => jsObject.prefix = v;
}

class File extends JsObjectWrapper<interop.FileJsImpl> implements _File {
  static final _expando = Expando<File>();

  factory File.getInstance(interop.FileJsImpl jsObject) {
    if (jsObject == null) return null;
    return _expando[jsObject] ??= File._fromJsObject(jsObject);
  }

  File._fromJsObject(interop.FileJsImpl jsObject)
      : super.fromJsObject(jsObject);

  @override
  Future<StorageResponse<File>> copy(
    dynamic /*String|Bucket|File*/ destination, [
    interop.CopyOptions options,
  ]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.copy(destination));
    } else {
      future = handleThenable(jsObject.copy(destination, options));
    }
    return _makeResponse(future, (o) => File.getInstance(o));
  }

  @override
  Future<Readable> createReadStream([interop.CreateReadStreamOptions options]) {
    if (options != null) {
      return handleThenable(jsObject.createReadStream(options));
    } else {
      return handleThenable(jsObject.createReadStream());
    }
  }

  @override
  Future<String> createResumableUpload([
    CreateResumableUploadOptions options,
  ]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.createResumableUpload(options.jsObject));
    } else {
      future = handleThenable(jsObject.createResumableUpload());
    }
    return future.then((value) => value.first as String);
  }

  @override
  Future<Writable> createWriteStream([CreateWriteStreamOptions options]) {
    if (options != null) {
      return handleThenable(jsObject.createWriteStream(options.jsObject));
    } else {
      return handleThenable(jsObject.createWriteStream());
    }
  }

  @override
  Future<Map<String, dynamic>> delete([interop.StorageBasicOptions options]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.delete(options));
    } else {
      future = handleThenable(jsObject.delete());
    }
    return future.then(_dartify);
  }

  @override
  void deleteResumableCache() => jsObject.deleteResumableCache();

  @override
  Future<Buffer> download([interop.CreateReadStreamOptions options]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.download(options));
    } else {
      future = handleThenable(jsObject.download());
    }
    return future.then((value) => value.first as Buffer);
  }

  @override
  Future<bool> exists([interop.StorageBasicOptions options]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.exists(options));
    } else {
      future = handleThenable(jsObject.exists());
    }
    return future.then((value) => value.first as bool);
  }
}

class GetBucketSignedUrlConfig
    extends JsObjectWrapper<interop.GetBucketSignedUrlConfigJsImpl> {
  static final _expando = Expando<GetBucketSignedUrlConfig>();

  factory GetBucketSignedUrlConfig({
    String action,
    String cname,
    dynamic /*String|num|JsDate*/ expires,
    Map<String, dynamic> extensionHeaders,
    Map<String, String> queryParams,
    String version,
    bool virtualHostedStyle,
  }) =>
      GetBucketSignedUrlConfig.getInstance(
          interop.GetBucketSignedUrlConfigJsImpl(
        action: action,
        cname: cname,
        expires: jsify(expires),
        extensionHeaders: jsify(extensionHeaders),
        queryParams: jsify(queryParams),
        version: version,
        virtualHostedStyle: virtualHostedStyle,
      ));

  factory GetBucketSignedUrlConfig.getInstance(
      interop.GetBucketSignedUrlConfigJsImpl jsObject) {
    if (jsObject == null) return null;
    return _expando[jsObject] ??=
        GetBucketSignedUrlConfig._fromJsObject(jsObject);
  }

  GetBucketSignedUrlConfig._fromJsObject(
      interop.GetBucketSignedUrlConfigJsImpl jsObject)
      : super.fromJsObject(jsObject);

  String get action => jsObject.action;
  set action(String v) => jsObject.action = v;

  String get cname => jsObject.cname;
  set cname(String v) => jsObject.cname = v;

  dynamic get expires => dartify(jsObject.expires);
  set expires(dynamic v) => jsObject.expires = jsify(v);

  Map<String, dynamic> get extensionHeaders =>
      dartify(jsObject.extensionHeaders);
  set extensionHeaders(Map<String, dynamic> v) =>
      jsObject.extensionHeaders = jsify(v);

  Map<String, String> get queryParams {
    final map = dartify(jsObject.queryParams);
    if (map == null) return null;
    return (map as Map<String, dynamic>).cast<String, String>();
  }

  set queryParams(Map<String, String> v) => jsObject.queryParams = jsify(v);

  String get version => jsObject.version;
  set version(String v) => jsObject.version = v;

  bool get virtualHostedStyle => jsObject.virtualHostedStyle;
  set virtualHostedStyle(bool v) => jsObject.virtualHostedStyle = v;
}

class LifeCycleRule extends JsObjectWrapper<interop.LifeCycleRuleJsImpl> {
  static final _expando = Expando<LifeCycleRule>();

  factory LifeCycleRule({
    interop.LifeCycleRuleAction action,
    Map<String, dynamic> condition,
    String storageClass,
  }) =>
      LifeCycleRule.getInstance(interop.LifeCycleRuleJsImpl(
        action: action,
        condition: jsify(condition),
        storageClass: storageClass,
      ));

  factory LifeCycleRule.getInstance(interop.LifeCycleRuleJsImpl jsObject) {
    if (jsObject == null) return null;
    return _expando[jsObject] ??= LifeCycleRule._fromJsObject(jsObject);
  }

  LifeCycleRule._fromJsObject(interop.LifeCycleRuleJsImpl jsObject)
      : super.fromJsObject(jsObject);
  interop.LifeCycleRuleAction get action => jsObject.action;
  set action(interop.LifeCycleRuleAction v) => jsObject.action = v;

  Map<String, dynamic> get condition => dartify(jsObject.condition);
  set condition(Map<String, dynamic> v) => jsObject.condition = jsify(v);

  String get storageClass => jsObject.storageClass;
  set storageClass(String v) => jsObject.storageClass = v;
}

class Storage extends JsObjectWrapper<interop.StorageJsImpl> {
  static final _expando = Expando<Storage>();

  factory Storage.getInstance(interop.StorageJsImpl jsObject) {
    if (jsObject == null) return null;
    return _expando[jsObject] ??= Storage._fromJsObject(jsObject);
  }

  Storage._fromJsObject(interop.StorageJsImpl jsObject)
      : super.fromJsObject(jsObject);

  Bucket bucket([String name]) {
    if (name != null) return Bucket.getInstance(jsObject.bucket(name));
    return Bucket.getInstance(jsObject.bucket());
  }
}

class StorageNotification extends JsObjectWrapper<interop.NotificationJsImpl> {
  static final _expando = Expando<StorageNotification>();

  factory StorageNotification.getInstance(interop.NotificationJsImpl jsObject) {
    if (jsObject == null) return null;
    return _expando[jsObject] ??= StorageNotification._fromJsObject(jsObject);
  }

  StorageNotification._fromJsObject(interop.NotificationJsImpl jsObject)
      : super.fromJsObject(jsObject);

  Future<Map<String, dynamic>> delete([interop.StorageBasicOptions options]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.delete(options));
    } else {
      future = handleThenable(jsObject.delete());
    }
    return future.then(_dartify);
  }

  Future<bool> exists(String topic, [Map<String, dynamic> options]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.exist(topic, jsify(options)));
    } else {
      future = handleThenable(jsObject.exist(topic));
    }
    return future.then((value) => value.first as bool);
  }

  Future<StorageResponse<StorageNotification>> get(
      [interop.GetBucketOptions options]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.get(options));
    } else {
      future = handleThenable(jsObject.get());
    }
    return _makeResponse(future, (o) => StorageNotification.getInstance(o));
  }

  Future<StorageResponse<Map<String, dynamic>>> getMetadata(
      [interop.StorageBasicOptions options]) {
    Future<List> future;
    if (options != null) {
      future = handleThenable(jsObject.getMetadata(options));
    } else {
      future = handleThenable(jsObject.getMetadata());
    }
    return _makeResponse(future, (o) => dartify(o) as Map<String, dynamic>);
  }
}

class StorageResponse<T> {
  final T data;
  final Map<String, dynamic> object;
  StorageResponse(this.data, this.object);
}

class StorageResponse0 {
  final Map<String, dynamic> apiResponse;
  StorageResponse0(this.apiResponse);
}

class UploadOptions extends JsObjectWrapper<interop.UploadOptionsJsImpl> {
  static final _expando = Expando<UploadOptions>();

  factory UploadOptions({
    String configPath,
    Map<String, dynamic> metadata,
    String origin,
    num offset,
    String predefinedAcl,
    bool private,
    bool public,
    String uri,
    String userProject,
    String destination,
    String encryptionKey,
    String kmsKeyName,
    bool resumable,
  }) =>
      UploadOptions.getInstance(interop.UploadOptionsJsImpl(
        configPath: configPath,
        metadata: jsify(metadata),
        origin: origin,
        offset: offset,
        predefinedAcl: predefinedAcl,
        private: private,
        public: public,
        uri: uri,
        userProject: userProject,
        destination: destination,
        encryptionKey: encryptionKey,
        kmsKeyName: kmsKeyName,
        resumable: resumable,
      ));

  factory UploadOptions.getInstance(interop.UploadOptionsJsImpl jsObject) {
    if (jsObject == null) return null;
    return _expando[jsObject] ??= UploadOptions._fromJsObject(jsObject);
  }

  UploadOptions._fromJsObject(interop.UploadOptionsJsImpl jsObject)
      : super.fromJsObject(jsObject);

  String get configPath => jsObject.configPath;
  set configPath(String v) => jsObject.configPath = v;

  String get destination => jsObject.destination;
  set destination(String v) => jsObject.destination = v;

  String get encryptionKey => jsObject.encryptionKey;
  set encryptionKey(String v) => jsObject.encryptionKey = v;

  String get kmsKeyName => jsObject.kmsKeyName;
  set kmsKeyName(String v) => jsObject.kmsKeyName = v;

  Map<String, dynamic> get metadata => dartify(jsObject.metadata);
  set metadata(Map<String, dynamic> v) => jsObject.metadata = jsify(v);

  num get offset => jsObject.offset;
  set offset(num v) => jsObject.offset = v;

  String get origin => jsObject.origin;
  set origin(String v) => jsObject.origin = v;

  String get predefinedAcl => jsObject.predefinedAcl;
  set predefinedAcl(String v) => jsObject.predefinedAcl = v;

  bool get private => jsObject.private;
  set private(bool v) => jsObject.private = v;

  bool get public => jsObject.public;
  set public(bool v) => jsObject.public = v;

  bool get resumable => jsObject.resumable;
  set resumable(bool v) => jsObject.resumable = v;

  String get uri => jsObject.uri;
  set uri(String v) => jsObject.uri = v;

  String get userProject => jsObject.userProject;
  set userProject(String v) => jsObject.userProject = v;
}

abstract class _Bucket {
  Future<Map<String, dynamic>> addLifecycleRule(
    LifeCycleRule rule, [
    interop.AddLifecycleRuleOptions options,
  ]);

  Future<StorageResponse<File>> combine(
    List /*String|File*/ sources,
    dynamic /*String,File*/ destination, [
    interop.CombineOptions options,
  ]);

  Future<StorageResponse<Bucket>> create([
    CreateBucketRequest metadata,
  ]);

  Future<StorageResponse<Channel>> createChannel(
    String id,
    interop.CreateChannelConfig config, [
    interop.StorageBasicOptions options,
  ]);

  Future<StorageResponse<StorageNotification>> createNotification(
    String topic, [
    CreateNotificationOptions options,
  ]);

  Future<Map<String, dynamic>> delete(
    interop.StorageBasicOptions options,
  );

  Future<void> deleteFiles([
    interop.DeleteFilesOptions options,
  ]);

  Future<Map<String, dynamic>> deleteLabels(List<String> labels);

  Future<Map<String, dynamic>> disableRequesterPays();

  Future<Map<String, dynamic>> enableLogging(EnableLoggingOptions options);

  Future<Map<String, dynamic>> enableRequesterPays();

  Future<bool> exists([interop.BucketExistsOptions options]);

  File file(String name, [interop.FileOptions options]);

  Future<StorageResponse<Bucket>> get([interop.GetBucketOptions options]);

  Future<List<File>> getFiles([interop.GetFilesOptions options]);

  Future<Map<String, dynamic>> getLabels([interop.StorageBasicOptions options]);

  Future<StorageResponse<Map<String, dynamic>>> getMetadata(
      [interop.StorageBasicOptions options]);

  Future<StorageResponse<List<StorageNotification>>> getNotifications(
      [interop.StorageBasicOptions options]);

  Future<String> getSignedUrl(GetBucketSignedUrlConfig config);

  Future<Map<String, dynamic>> lock(dynamic metageneration);

  Future<List<File>> makePrivate([interop.MakeBucketPrivateOptions options]);

  Future<List<File>> makePublic([interop.MakeBucketPublicOptions options]);

  StorageNotification notification(String id);

  Future<Map<String, dynamic>> removeRetentionPeriod();
  //TODO: implement setCorsConfiguration
  // Future setCorsConfiguration();

  Future<Map<String, dynamic>> setLabels(Map<String, String> labels,
      [interop.StorageBasicOptions options]);

  Future<Map<String, dynamic>> setMetadata(Map<String, String> metadata,
      [interop.StorageBasicOptions options]);

  Future<Map<String, dynamic>> setRetentionPeriod(int duration);

  Future<Map<String, dynamic>> setStorageClass(String storageClass,
      [interop.StorageBasicOptions options]);

  void setUserProject(String userProject);

  Future<StorageResponse<File>> upload(String path, [UploadOptions options]);
}

abstract class _CreateWriteStreamOptions {}

abstract class _File {
  Future<StorageResponse<File>> copy(dynamic destination,
      [interop.CopyOptions options]);

  Future<Readable> createReadStream([interop.CreateReadStreamOptions options]);

  Future<String> createResumableUpload([CreateResumableUploadOptions options]);

  Future<Writable> createWriteStream([CreateWriteStreamOptions options]);

  Future<Map<String, dynamic>> delete([interop.StorageBasicOptions options]);

  void deleteResumableCache();

  Future<Buffer> download([interop.CreateReadStreamOptions options]);

  Future<bool> exists([interop.StorageBasicOptions options]);
}
